import os
from os.path import exists, join, basename, abspath, isfile, isdir, relpath
import subprocess
import json

from utils import copy_file, clear_directory, ensure_directory, ensure_file_dir, copy_directory, get_all_files
import native.native_setup as native_setup
from make_config import MAKE_CONFIG, TOOLCHAIN_CONFIG, BaseConfig

CODE_OK = 0
CODE_FAILED_NO_GCC = 1001
CODE_FAILED_INVALID_MANIFEST = 1002
CODE_DUPLICATE_NAME = 1003
CODE_INVALID_JSON = 1004
CODE_INVALID_PATH = 1005

def prepare_compiler_executable(abi):
	from native.native_setup import abi_to_arch
	arch = abi_to_arch(abi)
	if arch is None:
		print(f"WARNING: Unregistered abi {abi}!")
	return native_setup.require_compiler_executable(arch=abi if arch is None else arch, install_if_required=True)

def get_manifest(directory):
	with open(join(directory, "manifest"), "r", encoding="utf-8") as file:
		manifest = json.load(file)
	return manifest

def get_name_from_manifest(directory):
	try:
		return get_manifest(directory)["shared"]["name"]
	except Exception:
		return None

def search_directory(parent, name):
	for dirpath, dirnames, filenames in os.walk(parent):
		for dir in dirnames:
			path = join(dirpath, dir)
			if get_name_from_manifest(path) == name:
				return path

def get_fake_so_dir(abi):
	fake_so_dir = TOOLCHAIN_CONFIG.get_path(join("toolchain/ndk/fakeso", abi))
	ensure_directory(fake_so_dir)
	return fake_so_dir

def add_fake_so(gcc, abi, name):
	file = join(get_fake_so_dir(abi), "lib" + name + ".so")
	if not isfile(file):
		result = subprocess.call([
			gcc, "-std=c++11",
			TOOLCHAIN_CONFIG.get_path("toolchain/bin/fakeso.cpp"),
			"-shared", "-o", file
		])
		print("Created fake so:", name, result, "OK" if result == CODE_OK else "ERROR")

def build_native_dir(directory, output_dir, cache_dir, abis, std_includes_path, rules: BaseConfig):
	executables = {}
	for abi in abis:
		executable = prepare_compiler_executable(abi)
		if executable is None:
			print("Failed to acquire GCC executable from NDK for abi", abi)
			return CODE_FAILED_NO_GCC
		executables[abi] = executable

	try:
		manifest = get_manifest(directory)
		targets = {}
		soname = "lib" + manifest["shared"]["name"] + ".so"
		for abi in abis:
			targets[abi] = join(output_dir, "so", abi, soname)
	except Exception as err:
		print("Failed to read manifest for directory", f"{directory}:", err)
		return CODE_FAILED_INVALID_MANIFEST

	keep_sources = rules.get_value("keepSources", fallback=False)
	if keep_sources:
		# copy everything and clear build files
		copy_directory(directory, output_dir, clear_dst=True)
		clear_directory(join(output_dir, "so"))
		os.remove(join(output_dir, soname))
	else:
		clear_directory(output_dir)

		# copy manifest
		copy_file(join(directory, "manifest"), join(output_dir, "manifest"))

		# copy includes
		keep_includes = rules.get_value("keepIncludes", fallback=True)
		for include_path in manifest["shared"]["include"]:
			src_include_path = join(directory, include_path)
			output_include_path = join(output_dir, include_path)
			if keep_includes:
				copy_directory(src_include_path, output_include_path, clear_dst=True)
			else:
				clear_directory(output_include_path)

	std_includes = []
	if exists(std_includes_path):
		for std_includes_dir in os.listdir(std_includes_path):
			std_includes_dirpath = join(std_includes_path, std_includes_dir)
			if isdir(std_includes_dirpath):
				std_includes.append(abspath(std_includes_dirpath))

	# compile for every abi
	overall_result = CODE_OK
	for abi in abis:
		print(f"* Compiling {basename(directory)} for {abi}")

		executable = executables[abi]
		gcc = [executable, "-std=c++11"]
		includes = []
		for std_includes_dir in std_includes:
			includes.append(f"-I{std_includes_dir}")
		dependencies = [f"-L{get_fake_so_dir(abi)}", "-landroid", "-lm", "-llog"]
		for link in rules.get_value("link", fallback=[]) + MAKE_CONFIG.get_value("linkNative", fallback=[]) + ["horizon"]:
			add_fake_so(executable, abi, link)
			dependencies.append(f"-l{link}")
		if "depends" in manifest:
			search_dir = abspath(join(directory, "..")) # always search for dependencies in current dir
			for dependency in manifest["depends"]:
				if dependency is not None:
					add_fake_so(executable, abi, dependency)
					dependencies.append("-l" + dependency)
					dependency_dir = search_directory(search_dir, dependency)
					if dependency_dir is not None:
						try:
							for include_dir in get_manifest(dependency_dir)["shared"]["include"]:
								includes.append("-I" + join(dependency_dir, include_dir))
						except KeyError:
							pass
				else:
					print(f"WARNING: Dependency directory {dependency} is not found, it will be skipped")

		# prepare directories
		source_files = get_all_files(directory, extensions=(".cpp", ".c"))
		preprocessed_dir = abspath(join(cache_dir, "preprocessed", abi))
		ensure_directory(preprocessed_dir)
		object_dir = abspath(join(cache_dir, "object", abi))
		ensure_directory(object_dir)

		# pre-process and compile changes
		import filecmp
		object_files = []
		recompiled_count = 0
		for file in source_files:
			relative_file = relpath(file, directory)
			print("Preprocessing " + relative_file + " " * 48, end="\r")

			object_file = join(object_dir, relative_file) + ".o"
			preprocessed_file = join(preprocessed_dir, relative_file)
			tmp_preprocessed_file = preprocessed_file + ".tmp"
			ensure_file_dir(preprocessed_file)
			ensure_file_dir(object_file)
			object_files.append(object_file)

			result = subprocess.call(gcc + [
				"-E", file, "-o", tmp_preprocessed_file
			] + includes)
			if result == CODE_OK:
				if not isfile(preprocessed_file) or not isfile(object_file) \
						or not filecmp.cmp(preprocessed_file, tmp_preprocessed_file):
					if isfile(preprocessed_file):
						os.remove(preprocessed_file)
					os.rename(tmp_preprocessed_file, preprocessed_file)
					if isfile(object_file):
						os.remove(object_file)

					print("Compiling " + relative_file + " " * 96)
					result = max(result, subprocess.call(gcc + [
						"-c", preprocessed_file, "-shared", "-o", object_file
					]))
					if result != CODE_OK:
						if isfile(object_file):
							os.remove(object_file)
						overall_result = result
					else:
						recompiled_count += 1
			else:
				if isfile(object_file):
					os.remove(object_file)
				overall_result = result

		if overall_result != CODE_OK:
			print("Failed to compile with result", overall_result)
			return overall_result
		else:
			print(f"Recompiled {recompiled_count}/{len(object_files)} files with result {overall_result}")

		ensure_file_dir(targets[abi])

		command = []
		command += gcc
		command += object_files
		command.append("-shared")
		command.append("-Wl,-soname=" + soname)
		command.append("-o")
		command.append(targets[abi])
		command += includes
		command += dependencies
		print("Linking object files")
		result = subprocess.call(command)
		if result == CODE_OK:
			print("Build successfully completed")
		else:
			print("Linker failed with result", result)
			overall_result = result
			return overall_result
		print()
	return overall_result

def compile_all_using_make_config(abis):
	from mod_structure import mod_structure
	import time
	start_time = time.time()

	std_includes = TOOLCHAIN_CONFIG.get_path("toolchain/stdincludes")
	if not exists(std_includes):
		print("\x1b[93mNot found toolchain/stdincludes, in most cases build will be failed, please install it via tasks.\x1b[0m")
	cache_dir = MAKE_CONFIG.get_build_path("gcc")
	ensure_directory(cache_dir)
	mod_structure.cleanup_build_target("native")

	overall_result = CODE_OK

	for native_dir in MAKE_CONFIG.get_filtered_list("compile", "type", ("native")):
		if "source" not in native_dir:
			print("Skipped invalid native directory json", native_dir)
			overall_result = CODE_INVALID_JSON
			continue
		for native_dir_path in MAKE_CONFIG.get_paths(native_dir["source"]):
			if isdir(native_dir_path):
				directory_name = basename(native_dir_path)
				result = build_native_dir(
					native_dir_path,
					mod_structure.new_build_target("native", directory_name + "{}"),
					join(cache_dir, directory_name),
					abis,
					std_includes,
					BaseConfig(native_dir["rules"] if "rules" in native_dir else {})
				)
				if result == CODE_FAILED_NO_GCC:
					return overall_result
				if result != CODE_OK:
					overall_result = result
			else:
				print("Skipped non existing native directory", native_dir["source"])
				overall_result = CODE_INVALID_PATH

	mod_structure.update_build_config_list("nativeDirs")
	print(f"Completed native build in {int((time.time() - start_time) * 100) / 100}s with result {overall_result} - {'OK' if overall_result == CODE_OK else 'ERROR'}")
	return overall_result


if __name__ == "__main__":
	compile_all_using_make_config(["x86"])



import hashlib
import os
from os.path import join, basename, abspath, exists, isfile, isdir
import platform
import json

from base_config import BaseConfig

class MakeConfig(BaseConfig):
	def __init__(self, filename, base = None):
		if not isfile(filename):
			from task import error
			error(f"Not found {basename(filename)}, are you sure that selected project exists?")
		self.filename = filename
		self.root_dir = abspath(join(self.filename, ".."))
		with open(filename, encoding="utf-8") as file:
			self.json = json.load(file)
		if base is not None:
			if "global" in self.json:
				globalconfig = self.get_value("global", {})
				for entry in globalconfig:
					self.set_value(entry, globalconfig[entry])
				self.remove_value("global")
				self.save()
			if "make" in self.json:
				makeconfig = self.get_value("make", {})
				if "linkNative" in makeconfig:
					self.set_value("linkNative", makeconfig["linkNative"])
				if "excludeFromRelease" in makeconfig:
					self.set_value("excludeFromRelease", makeconfig["excludeFromRelease"])
				self.remove_value("make")
				self.save()
		BaseConfig.__init__(self, self.json, base)

	def save(self):
		with open(self.filename, "w", encoding="utf-8") as make_file:
			make_file.write(json.dumps(self.json, indent="\t") + "\n")

	def get_absolute_path(self, path):
		relative_path = self.get_path(path)
		return relative_path if exists(relative_path) else abspath(path) if exists(abspath(path)) else relative_path

	def get_path(self, relative_path):
		return abspath(join(self.root_dir, relative_path))

	def get_paths(self, relative_path, filter = None, paths = None):
		if paths is None:
			paths = []
		if len(relative_path) > 0 and relative_path[-1] == "*":
			path = self.get_path(relative_path[:-1])
			if not isdir(path):
				return []
			for filename in os.listdir(path):
				file = join(path, filename)
				if filter is None or filter(file):
					paths.append(file)
		else:
			path = self.get_path(relative_path)
			if filter is None or filter(path):
				paths.append(path)
		return paths

	def get_adb(self):
		try:
			import shutil
			if shutil.which("adb") is not None:
				return "adb"
		except:
			pass
		if platform.system() == "Windows":
			return self.get_path("toolchain/adb/adb.exe")
		return self.get_path("toolchain/adb/adb")

class ToolchainMakeConfig(MakeConfig):
	def __init__(self, filename):
		prototype = MakeConfig(filename)
		self.current_project = prototype.get_value("currentProject")
		if self.current_project is not None:
			make_path = prototype.get_absolute_path(self.current_project + "/make.json")
			if isfile(make_path):
				MakeConfig.__init__(self, make_path, MakeConfig(filename))
				self.project_unique_name = self.unique_folder_name(self.root_dir)
				return
			self.current_project = None
		MakeConfig.__init__(self, filename)

	def assure_project_selected(self):
		if self.current_project is None:
			from task import registered_tasks
			print("Not found any opened project, nothing to do.")
			registered_tasks["selectProject"]()
			return MAKE_CONFIG.assure_project_selected()

	def get_path(self, relative_path):
		self.assure_project_selected()
		return MakeConfig.get_path(self, relative_path)

	def get_absolute_path(self, path):
		self.assure_project_selected()
		return MakeConfig.get_absolute_path(self, path)

	def get_paths(self, relative_path, filter = None, paths = None):
		self.assure_project_selected()
		return MakeConfig.get_paths(self, relative_path, filter, paths)

	def get_build_path(self, relative_path):
		self.assure_project_selected()
		return self.prototype.get_path(join(
			"toolchain", "build", self.project_unique_name, relative_path
		))

	@staticmethod
	def unique_folder_name(path):
		md5 = hashlib.md5()
		md5.update(bytes(path, "utf-8"))
		return basename(path) + "-" + md5.hexdigest()


# search for toolchain.json
MAKE_CONFIG = None
for i in range(0, 4):
	make_file = join("../" * i, "toolchain.json")
	if isfile(make_file):
		MAKE_CONFIG = ToolchainMakeConfig(make_file)
		break
if MAKE_CONFIG is None:
	raise RuntimeError("Not found toolchain.json!")
TOOLCHAIN_CONFIG = MakeConfig(MAKE_CONFIG.filename) \
	if MAKE_CONFIG.current_project is None else MAKE_CONFIG.prototype



from os.path import isfile, abspath, relpath, join
import json

from make_config import MakeConfig, MAKE_CONFIG, TOOLCHAIN_CONFIG
from base_config import BaseConfig

class WorkspaceNotAvailable(RuntimeError):
	def __init__(self, *args):
		RuntimeError.__init__(self, "Workspace is not available", *args)

class CodeWorkspace(BaseConfig):
	def __init__(self, filename):
		if not isfile(filename):
			return BaseConfig.__init__(self, {})
		self.filename = filename
		self.root_dir = abspath(join(self.filename, ".."))
		with open(filename, encoding="utf-8") as file:
			self.json = json.load(file)
		BaseConfig.__init__(self, self.json)

	def available(self):
		return hasattr(self, "filename") and isfile(self.filename)

	def get_path(self, relative_path):
		if not self.available():
			raise WorkspaceNotAvailable()
		return MakeConfig.get_path(self, relative_path)

	def get_toolchain_path(self, relative_path = ""):
		if not self.available():
			raise WorkspaceNotAvailable()
		return relpath(TOOLCHAIN_CONFIG.get_path(relative_path), self.root_dir)

	def save(self):
		if not self.available():
			raise WorkspaceNotAvailable()
		return MakeConfig.save(self)


CODE_WORKSPACE = CodeWorkspace(TOOLCHAIN_CONFIG.get_absolute_path(MAKE_CONFIG.get_value("workspaceFile")))
CODE_SETTINGS = CodeWorkspace(TOOLCHAIN_CONFIG.get_path(".vscode/settings.json"))



import os
from os.path import join, isfile, isdir
import shutil
import sys
from urllib import request
from urllib.error import URLError

from make_config import TOOLCHAIN_CONFIG
from utils import ensure_file_dir, ensure_not_whitespace, AttributeZipFile
from shell import *

class Component():
	def __init__(self, keyword, name, location = "", packurl = None, commiturl = None, branch = None):
		self.keyword = keyword
		self.name = name
		self.location = location
		if branch is not None:
			self.packurl = "https://codeload.github.com/zheka2304/innercore-mod-toolchain/zip/" + branch
			self.commiturl = "https://raw.githubusercontent.com/zheka2304/innercore-mod-toolchain/" + branch + "/.commit"
			self.branch = branch
		if packurl is not None:
			self.packurl = packurl
		if commiturl is not None:
			self.commiturl = commiturl

COMPONENTS = {
	"adb": Component("adb", "Android Debug Bridge", "toolchain/adb", branch="adb"),
	"declarations": Component("declarations", "TypeScript Declarations", "toolchain/declarations", branch="includes"),
	"java": Component("java", "Java R8/D8 Compiler", "toolchain/bin/r8", branch="r8"),
	"classpath": Component("classpath", "Java Classpath", "toolchain/classpath", branch="classpath"),
	"cpp": Component("cpp", "C++ GCC Compiler (NDK)", "toolchain/ndk"),
	"stdincludes": Component("stdincludes", "C++ Headers", "toolchain/stdincludes", branch="stdincludes")
}

def put_components(installed = []):
	return [
		Switch("component:" + key, COMPONENTS[key].name, True if key in installed else False) for key in COMPONENTS
	]

def resolve_components(interactables):
	keywords = []
	for interactable in interactables:
		if interactable.key is not None and interactable.key.startswith("component:") and interactable.checked:
			keywords.append(interactable.key.partition(":")[2])
	return keywords

def which_installed():
	installed = []
	for componentname in COMPONENTS:
		component = COMPONENTS[componentname]
		path = TOOLCHAIN_CONFIG.get_path(component.location)
		if not isdir(path):
			continue
		if component.keyword == "cpp":
			installed.append("cpp")
			continue
		if isfile(join(path, ".commit")) or TOOLCHAIN_CONFIG.get_value("componentInstallationWithoutCommit", False):
			installed.append(component.keyword)
	return installed

def download_component(component, shell, progress):
	if not hasattr(component, "packurl") or component.packurl is None:
		progress.seek(0, f"Component {component.keyword} packurl property must be defined!")
		shell.render()
		return 1
	path = TOOLCHAIN_CONFIG.get_path(f"toolchain/temp/{component.keyword}.zip")
	ensure_file_dir(path)
	if isfile(path):
		return 0
	with request.urlopen(component.packurl) as response:
		with open(path, "wb") as f:
			downloaded = 0
			while True:
				buffer = response.read(8192)
				if not buffer:
					break
				downloaded += len(buffer)
				progress.seek(0.5, f"Downloading ({(downloaded / 1048576):.1f}MiB)")
				shell.render()
				f.write(buffer)
	progress.seek(1, f"Downloaded {(downloaded / 1048576):.1f}MiB")
	shell.render()
	return 0

def extract_component(component, shell, progress):
	temp = TOOLCHAIN_CONFIG.get_path("toolchain/temp")
	archive_path = join(temp, component.keyword + ".zip")
	if not isfile(archive_path):
		progress.seek(0, f"Component {component.keyword} downloaded nothing to extract!")
		shell.render()
		return 1
	progress.seek(0.5, f"Extracting to {component.location}")
	extract_to = temp if hasattr(component, "branch") else join(temp, component.keyword)
	with AttributeZipFile(archive_path, "r") as archive:
		archive.extractall(extract_to)
	if hasattr(component, "branch"):
		extract_to = join(extract_to, "innercore-mod-toolchain-" + component.branch)
	if not isdir(extract_to):
		progress.seek(0, f"Component {component.keyword} does not contain any content!")
		shell.render()
		return 2
	output = TOOLCHAIN_CONFIG.get_path(component.location)
	if isdir(output):
		shutil.rmtree(output, ignore_errors=True)
	elif isfile(output):
		shutil.move(output, output + ".bak")
	os.makedirs(output, exist_ok=True)
	shutil.copytree(extract_to, output, dirs_exist_ok=True)
	progress.seek(1, "Cleaning up")
	shell.render()
	shutil.rmtree(extract_to, ignore_errors=True)
	os.remove(archive_path)
	return 0

def install_components(components):
	if len(components) == 0:
		return
	shell = InteractiveShell(lines_per_page=max(len(components), 9))
	shell.enter()
	for componentname in components:
		if not componentname in COMPONENTS:
			print(f"Not found component {componentname}!")
			continue
		if componentname == "cpp":
			continue
		component = COMPONENTS[componentname]
		progress = Progress(text=component.name)
		shell.interactables.append(progress)
		shell.render()
		if fetch_component(component):
			progress.seek(1)
			shell.render()
			continue
		try:
			if download_component(component, shell, progress) == 0:
				if extract_component(component, shell, progress) == 0:
					progress.seek(1, component.name)
		except URLError:
			continue
		except BaseException as err:
			progress.seek(0, f"{component.keyword}: {err}")
		shell.render()
	if "cpp" in components:
		abis = TOOLCHAIN_CONFIG.get_value("abis", [])
		if not isinstance(abis, list):
			abis = []
		abi = TOOLCHAIN_CONFIG.get_value("debugAbi")
		if abi is None and len(abis) == 0:
			from task import error
			error("Please describe options 'abis' or 'debugAbi' in your toolchain.json before install NDK!")
		if abi is not None and not abi in abis:
			abis.append(abi)
		from native.native_setup import abi_to_arch, check_installed, install
		for arch in abis:
			if not check_installed(abi_to_arch(arch)):
				install(abi_to_arch(arch), reinstall=True)
	shell.interactables.append(Interrupt())

def fetch_component(component):
	output = TOOLCHAIN_CONFIG.get_path(component.location)
	if component.keyword == "cpp":
		return isdir(output)
	if isdir(output):
		if TOOLCHAIN_CONFIG.get_value("componentInstallationWithoutCommit", False):
			return True
		if not isfile(join(output, ".commit")):
			return False
	else:
		return False
	if not hasattr(component, "commiturl"):
		return True
	try:
		with open(join(output, ".commit")) as commit_file:
			response = request.urlopen(component.commiturl)
			return perform_diff(response.read().decode("utf-8"), commit_file.read())
	except BaseException:
		return True

def perform_diff(a, b):
	return str(a).strip() == str(b).strip()

def fetch_components():
	upgradable = []
	for componentname in which_installed():
		if not componentname in COMPONENTS:
			print(f"Not found component {componentname}!")
			continue
		if not fetch_component(COMPONENTS[componentname]):
			upgradable.append(componentname)
	return upgradable

def get_username():
	try:
		return os.environ["USER"]
	except KeyError:
		return None

def startup():
	print("Welcome to Inner Core Mod Toolchain!", end="")
	shell = SelectiveShell()
	shell.interactables += [
		Separator(),
		Notice("Today we've complete your distribution installation."),
		Notice("Just let realize some things before downloading, and"),
		Notice("modding will be started in a few moments."),
		Separator(),
		Progress(progress=0.25, text=" " + "Howdy!".center(45) + ">")
	]
	shell.interactables += [
		Separator(),
		Input("user", "I'll will be ", template=get_username()),
		Notice("Author name identifies you, it will be used as default"),
		Notice("`author` property when you're starting new project."),
		Separator(),
		Progress(progress=0.5, text="<" + "Who are you?".center(45) + ">")
	]
	preffered = which_installed()
	if not "declarations" in preffered:
		preffered.append("declarations")
	if not "java" in preffered:
		preffered.append("java")
	try:
		import shutil
		if shutil.which("adb") is None and not "adb" in preffered:
			preffered.append("adb")
	except BaseException:
		pass
	components = put_components(preffered)
	interactables = [
		Notice("Which components will be installed?")
	]
	component = 0
	index = len(interactables)
	while True:
		if index % shell.lines_per_page == shell.lines_per_page - 1:
			interactables.append(Progress(progress=0.75, text="<" + "Configure your toolchain".center(45) + (
				">" if component < len(components) + 3 else "+"
			)))
			if component == len(components) + 3:
				break
		elif component < len(components) + 3:
			if component < len(components):
				interactables.append(components[component])
			elif component == len(components):
				interactables.append(Separator())
			elif component == len(components) + 1:
				interactables.append(Notice("Any of components above may be installed when it might"))
			elif component == len(components) + 2:
				interactables.append(Notice("be required. Selected will be availabled right now."))
			component += 1
		else:
			required_lines = shell.lines_per_page - ((index + 1) % shell.lines_per_page)
			interactables.append(Separator(size=required_lines))
			index += required_lines - 1
		index += 1
	shell.interactables += [
		*interactables,
		Interrupt()
	]
	try:
		shell.loop()
	except KeyboardInterrupt:
		shell.leave()
		return print()
	print()
	username = shell.get_interactable("user").read()
	if ensure_not_whitespace(username) is not None:
		print("Who are you?", f"\x1b[2m{username}\x1b[0m")
		TOOLCHAIN_CONFIG.set_value("template.author", username)
		TOOLCHAIN_CONFIG.save()
	installation = resolve_components(shell.interactables)
	if len(installation) > 0:
		print("Which components will be installed? ", "\x1b[2m", ", ".join(installation), "\x1b[0m", sep="")
		install_components(installation)

def foreign():
	print("Which components will be upgraded?", end="")
	shell = SelectiveShell(lines_per_page=min(len(COMPONENTS), 9))
	shell.interactables += put_components(which_installed())
	shell.interactables.append(Interrupt())
	try:
		shell.loop()
	except KeyboardInterrupt:
		shell.leave()
		print()
		return 0
	installation = resolve_components(shell.interactables)
	if len(installation) > 0:
		print("Which components will be upgraded? ", "\x1b[2m", ", ".join(installation), "\x1b[0m", sep="")
		install_components(installation)
	else:
		print("Nothing to perform.")
	return 0


if __name__ == "__main__":
	if "--help" in sys.argv:
		print("Usage: component.py <options> [components]")
		print(" " * 2 + "--startup: Startup configuration instead of component installation.")
		exit(0)
	if "--startup" in sys.argv or "-s" in sys.argv:
		startup()
	else:
		foreign()



class BaseConfig:
	def __init__(self, json, base = None):
		self.json = json
		self.prototype = base

	def get_value(self, name, fallback = None, accept_prototype = True):
		rawname = name.split(".")
		value = self.json
		while len(rawname) > 0 and len(rawname[0]) > 0:
			key = rawname.pop(0)
			if key in value:
				value = value[key]
			elif accept_prototype and self.prototype is not None:
				return self.prototype.get_value(name, fallback)
			else:
				return fallback
		return value

	def set_value(self, name, what):
		rawname = name.split(".")
		value = self.json
		while len(rawname) > 1 and len(rawname[0]) > 0:
			key = rawname.pop(0)
			if not key in value:
				value[key] = {}
			value = value[key]
		if len(rawname[0]) > 0:
			value[rawname.pop()] = what

	def remove_value(self, name):
		rawname = name.split(".")
		value = self.json
		while len(rawname) > 1 and len(rawname[0]) > 0:
			key = rawname.pop(0)
			if not key in value:
				return
			value = value[key]
		if len(rawname[0]) > 0:
			del value[rawname.pop()]
			if value != self.json and len(value) == 0:
				self.remove_value(name.rsplit(".", 1)[0])

	def get_config(self, name, not_none = False):
		value = self.get_value(name)
		if isinstance(value, dict):
			return BaseConfig(value)
		else:
			return BaseConfig({}) if not_none else None

	def get_filtered_list(self, name, prop, values):
		value = self.get_value(name)
		filtered = []
		if isinstance(value, list):
			for obj in value:
				if isinstance(obj, dict) and prop in obj and obj[prop] in values:
					filtered.append(obj)
		return filtered
